structure FRiver :> FRiver =
  (* Hank Dvid kdja alapjn *)
struct

open TRiver;

(* A bemenetet, ill. a kimenetet a terminlra irnythatjuk Linux alatt
   a /dev/stdin, ill. a /dev/stdout, MS-DOS/Windows alatt mindkt esetben
   a con fjlnv megadsval.
*)

exception FRiver of string;

(* valid_problem rm = rm is a valid problem specification; the first argument
                      of River.river is guaranteed to satisfy this test.
 *)
fun valid_problem (Bsize, FieldL) =
    let
        val (MaxR, MaxC) = Bsize
			   
        fun valid_location (Row, Column) = 1 <= Row andalso 1 <= Column andalso
					   Row <= MaxR andalso Column <= MaxC
							       
        fun valid_type (Port n) = 1 <= n andalso n <= 7
          | valid_type _ = true
     
        fun valid_field (Row, Column, Type) =
              valid_location (Row, Column) andalso valid_type Type
     
        fun distance (SOME (S1, O1), SOME (S2, O2)) =
              Math.sqrt(Math.pow(real(S1)-real(S2),2.0)+
              Math.pow(real(O1)-real(O2),2.0))
          | distance _ = 10e6
     
        fun lex_compare ((x1,y1,_), (x2,y2,_)) =
            if      x1 < x2 orelse x1 = x2 andalso y1 < y2
            then LESS
            else if x1 > x2 orelse x1 = x2 andalso y1 > y2
            then GREATER
            else EQUAL
     
        fun neighbors_different (rp1::(rps as rp2::_)) =
              lex_compare (rp1, rp2) <> EQUAL
              andalso neighbors_different rps
          | neighbors_different _ = true
     
        fun source (_, _, Source) = true
          | source _ = false
     
        fun lake (_, _, Lake) = true
          | lake _ = false
     
        fun coords (SOME (S, O, _)) = SOME (S, O)
          | coords NONE = NONE
    in
        (* the size of the board is correct: *)
        1 <= MaxR andalso 1 <= MaxC andalso
	
        (* all elements of FieldL are valid: *)
        List.all valid_field FieldL andalso
     
        (* source and lake are not side neighbors: *)
        distance(coords(List.find source FieldL),
                 coords(List.find lake FieldL)) > 1.0 andalso
                 
        (* there is at most one source in the list: *)
        List.length (List.filter source FieldL) < 2 andalso
        
        (* there is at most one lake in the list: *)
        List.length (List.filter lake FieldL) < 2 andalso

        (* FieldL is lexicographically sorted: *)
        Listsort.sorted lex_compare FieldL andalso

        (* all elements of FieldL specify a field with a unique coordinate: *)
        neighbors_different FieldL
    end
 
(* river_in f = Az f szvegfjlbl beolvasott adatokbl kpzett folytrkp
   river_in : string -> TRiver.rivermap
*)
fun river_in f =
      let
          fun string2int s =
                let
                    val i = Int.fromString s
                in 
                    if isSome i
                    then valOf i
                    else raise FRiver("Syntactic error in file " ^ f)
                end

          fun par [a,b] = (a,b)
            | par _ = raise FRiver("Syntactic error in file " ^ f)

          fun tipus ["folyo"]  = River
            | tipus ["source"] = Source
            | tipus ["lake"]  = Lake
            | tipus ["ret"]  = Meadow
            | tipus ["kikoto",n] = Port (string2int n)
            | tipus _    = raise FRiver("Syntactic error in file " ^ f)

          fun mezo (a::b::c) = (string2int a,string2int b,tipus c)
            | mezo _ = raise FRiver("Syntactic error in file " ^ f)
 
          val tokenizedLine = (String.tokens Char.isSpace) o TextIO.inputLine
 
          fun mezol_be s = if TextIO.endOfStream s
                           then []
                           else (mezo o tokenizedLine) s :: mezol_be s
 
          val istr = TextIO.openIn f

          val meret = (par o (List.map string2int) o tokenizedLine) istr

          val mezol = mezol_be istr

          val terkep = (meret, mezol) before TextIO.closeIn istr
      in
         if valid_problem terkep
         then terkep
         else raise FRiver("Erroneous rivermap in file " ^ f)
      end

(* folyorajz (ft, fv) = a fv kgyvonalnak megfelel folyrajz az ft
                        folytrkpen
   folyorajz : TRiver.rivermap * TRiver.riverpath -> string
*)
fun folyorajz (((m,sz), mzk), fv) =
      let
          fun paircmp ((a,b), (c,d)) =
                case Int.compare(a,c) of
                     EQUAL => Int.compare(b,d)
                   | other => other

          fun displayChar River  = #"@"
            | displayChar Source  = #"F"
            | displayChar Lake   = #"T"
            | displayChar Meadow   = #"R"
            | displayChar (Port n) = chr(ord #"0" + n)

          fun insert ((S,O,c), bmap) = Binarymap.insert(bmap, (S,O), c)

          val pk = Binarymap.mkDict paircmp
          val pk = foldl insert pk (map (fn (S,O) => (S,O,#"@")) fv)
          val pk = foldl insert pk
                         (map (fn (S,O,v) => (S,O,displayChar v)) mzk)

          val tab   = String.concat o List.tabulate
          val vonal = tab (sz, fn _ => "+---") ^ "+\n"

          fun cellarajz S O = case Binarymap.peek(pk, (S+1,O+1)) of
                 SOME c => " " ^ str c ^ " |"
               | NONE   => "   |"

          fun sorrajz i = "|" ^ tab (sz, cellarajz i) ^ "\n" ^ vonal
      in
          vonal ^ tab (m, sorrajz)
      end

(* folyorajzok (ft, fvk) = az fvk megoldslistnak megfelel folyrajzok az
                           ft folytrkpen, res sorokkal elvlasztva
   folyorajzok : TRiver.rivermap * TRiver.riverpath list -> string
*)
fun folyorajzok (ft, fvk) =
      foldl (fn (fv,s) => s ^ folyorajz(ft, fv) ^ "\n") "" fvk

(* river_out(m, ft, fvk) = A kt folytrkp fvk megoldslistjt olvashat
                          formban kirja az m szvegfjlba.
   river_out : string * TRiver.rivermap * TRiver.riverpath list -> unit
*)
fun river_out(m, ft, fvk) =
      let
          val ostr = TextIO.openOut m
          val _ = TextIO.output(ostr, folyorajzok(ft, fvk))
      in
         TextIO.closeOut ostr
      end

(* solve(f, m) = Beolvas egy feladvnyt az f szvegfjlbl s
                  kirja az sszes megoldst az m szvegfjlba. Ehhez
                  termszetesen szksge van a River.river fggvnyre.
                  Eredmnye az f nevet, a megoldsok szmt s a futsi
                  idt mp-ben tartalmaz fzr.
   solve : string * string -> string
*)
fun solve (f, m) =
      let
          val ft = river_in f
          val start = Timer.startCPUTimer()
          val fvk = River.river ft
          val {usr=tim,...} = Timer.checkCPUTimer start
          val _ = river_out(m, ft, fvk)
      in
         "Problem: " ^ f ^ ", " ^
         Int.toString(length fvk) ^ " solution(s), " ^
         "time: " ^ Time.fmt 2 tim ^ " sec\n"
      end

end
